@@ -34,7 +34,7 @@ GNUTLS_Version GNUTLS_TLS1 = {0, 3, 1};
 GNUTLS_Version GNUTLS_SSL3 = {0, 3, 0};
 
 static gnutls_version_entry sup_versions[] = {
-	{ "SSL3", &GNUTLS_SSL3,  0 },
+	{ "SSL3", &GNUTLS_SSL3,  1 },
 	{ "TLS1", &GNUTLS_TLS1,  1 },
 	{0}
 };
@@ -47,8 +47,8 @@ static gnutls_version_entry sup_versions[] = {
                         GNUTLS_VERSION_LOOP( if( (p->id->local == version.local)&&(p->id->major == version.major)&&(p->id->minor == version.minor) ) { a; break; } )
 
 
-#define GNUTLS_CIPHER_ENTRY(name, blksize, keysize, block, iv, priority) \
-	{ #name, name, blksize, keysize, block, iv, priority }
+#define GNUTLS_CIPHER_ENTRY(name, blksize, keysize, block, iv) \
+	{ #name, name, blksize, keysize, block, iv }
 
 struct gnutls_cipher_entry {
 	char *name;
@@ -57,21 +57,18 @@ struct gnutls_cipher_entry {
 	size_t keysize;
 	size_t block;
 	size_t iv;
-	int priority;
 };
 typedef struct gnutls_cipher_entry gnutls_cipher_entry;
 
 static gnutls_cipher_entry algorithms[] = {
-	GNUTLS_CIPHER_ENTRY(GNUTLS_3DES, 8, 24, 1, 8, -1),
-	GNUTLS_CIPHER_ENTRY(GNUTLS_AES, 16, 16, 1, 16, -1),
+	GNUTLS_CIPHER_ENTRY(GNUTLS_3DES, 8, 24, 1, 8),
+	GNUTLS_CIPHER_ENTRY(GNUTLS_RIJNDAEL, 16, 16, 1, 16),
 #ifdef USE_MCRYPT
-	GNUTLS_CIPHER_ENTRY(GNUTLS_DES, 8, 8, 1, 8, -1),
-	GNUTLS_CIPHER_ENTRY(GNUTLS_ARCFOUR, 1, 16, 0, 0, -1),
+	GNUTLS_CIPHER_ENTRY(GNUTLS_ARCFOUR, 1, 16, 0, 0,),
 #else
-	GNUTLS_CIPHER_ENTRY(GNUTLS_DES, 8, 8, 1, 8, -2),
-	GNUTLS_CIPHER_ENTRY(GNUTLS_ARCFOUR, 1, 16, 0, 0, -2),
+	GNUTLS_CIPHER_ENTRY(GNUTLS_ARCFOUR, 1, 16, 0, 0),
 #endif
-	GNUTLS_CIPHER_ENTRY(GNUTLS_NULL, 1, 0, 0, 0, -1),
+	GNUTLS_CIPHER_ENTRY(GNUTLS_NULL, 1, 0, 0, 0),
 	{0}
 };
 
@@ -83,21 +80,20 @@ static gnutls_cipher_entry algorithms[] = {
                         GNUTLS_LOOP( if(p->id == algorithm) { a; break; } )
 
 
-#define GNUTLS_HASH_ENTRY(name, hashsize, priority) \
-	{ #name, name, hashsize, priority }
+#define GNUTLS_HASH_ENTRY(name, hashsize) \
+	{ #name, name, hashsize }
 
 struct gnutls_hash_entry {
 	char *name;
 	MACAlgorithm id;
 	size_t digestsize;
-	int priority;
 };
 typedef struct gnutls_hash_entry gnutls_hash_entry;
 
 static gnutls_hash_entry hash_algorithms[] = {
-	GNUTLS_HASH_ENTRY(GNUTLS_MAC_SHA, 20, -1),
-	GNUTLS_HASH_ENTRY(GNUTLS_MAC_MD5, 16, -1),
-	GNUTLS_HASH_ENTRY(GNUTLS_MAC_NULL, 0, -1),
+	GNUTLS_HASH_ENTRY(GNUTLS_MAC_SHA, 20),
+	GNUTLS_HASH_ENTRY(GNUTLS_MAC_MD5, 16),
+	GNUTLS_HASH_ENTRY(GNUTLS_MAC_NULL, 0),
 	{0}
 };
 
@@ -109,10 +105,34 @@ static gnutls_hash_entry hash_algorithms[] = {
                         GNUTLS_HASH_LOOP( if(p->id == algorithm) { a; break; } )
 
 
+/* Compression Section */
+#define GNUTLS_COMPRESSION_ENTRY(name) \
+	{ #name, name }
+
+struct gnutls_compression_entry {
+	char *name;
+	CompressionMethod id;
+};
+
+typedef struct gnutls_compression_entry gnutls_compression_entry;
+static gnutls_compression_entry compression_algorithms[] = {
+	GNUTLS_COMPRESSION_ENTRY(GNUTLS_COMPRESSION_NULL),
+//#ifdef HAVE_LIBZ
+	GNUTLS_COMPRESSION_ENTRY(GNUTLS_ZLIB),
+//#endif
+	{0}
+};
+
+#define GNUTLS_COMPRESSION_LOOP(b) \
+        gnutls_compression_entry *p; \
+                for(p = compression_algorithms; p->name != NULL; p++) { b ; }
+#define GNUTLS_COMPRESSION_ALG_LOOP(a) \
+                        GNUTLS_COMPRESSION_LOOP( if(p->id == algorithm) { a; break; } )
 
 
-#define GNUTLS_KX_ALGO_ENTRY(name, server_cert, server_kx, client_cert, RSA_premaster, DH_public_value, priority) \
-	{ #name, name, server_cert, server_kx, client_cert, RSA_premaster, DH_public_value, priority }
+/* Key Exchange Section */
+#define GNUTLS_KX_ALGO_ENTRY(name, server_cert, server_kx, client_cert, RSA_premaster, DH_public_value) \
+	{ #name, name, server_cert, server_kx, client_cert, RSA_premaster, DH_public_value }
 
 struct gnutls_kx_algo_entry {
 	char *name;
@@ -122,17 +142,16 @@ struct gnutls_kx_algo_entry {
 	int client_cert;
 	int RSA_premaster;
 	int DH_public_value;
-	int priority;
 };
 typedef struct gnutls_kx_algo_entry gnutls_kx_algo_entry;
 
 static gnutls_kx_algo_entry kx_algorithms[] = {
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_ANON_DH, 0, 1, 0, 0, 1, -1),
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_RSA, 1, 0, 1, 1, 0, -2),
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DHE_DSS, 1, 1, 1, 0, 0, -1),
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DHE_RSA, 1, 1, 1, 0, 0, -1),
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DH_DSS, 1, 0, 1, 0, 0, -2),
-	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DH_RSA, 1, 0, 1, 0, 0, -2),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_ANON_DH, 0, 1, 0, 0, 1),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_RSA, 1, 0, 1, 1, 0),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DHE_DSS, 1, 1, 1, 0, 0),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DHE_RSA, 1, 1, 1, 0, 0),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DH_DSS, 1, 0, 1, 0, 0),
+	GNUTLS_KX_ALGO_ENTRY(GNUTLS_KX_DH_RSA, 1, 0, 1, 0, 0),
 	{0}
 };
 
@@ -146,8 +165,8 @@ static gnutls_kx_algo_entry kx_algorithms[] = {
 
 
 /* Cipher SUITES */
-#define GNUTLS_CIPHER_SUITE_ENTRY( name, block_algorithm, kx_algorithm, mac_algorithm) \
-	{ #name, {name}, block_algorithm, kx_algorithm, mac_algorithm }
+#define GNUTLS_CIPHER_SUITE_ENTRY( name, block_algorithm, kx_algorithm, mac_algorithm, compression_algorithm) \
+	{ #name, {name}, block_algorithm, kx_algorithm, mac_algorithm, compression_algorithm }
 
 typedef struct {
 	char *name;
@@ -155,10 +174,12 @@ typedef struct {
 	BulkCipherAlgorithm block_algorithm;
 	KXAlgorithm kx_algorithm;
 	MACAlgorithm mac_algorithm;
+	CompressionMethod compression_algorithm;
 } gnutls_cipher_suite_entry;
 
 #define GNUTLS_DH_anon_WITH_3DES_EDE_CBC_SHA { 0x00, 0x1B }
 #define GNUTLS_DH_anon_WITH_ARCFOUR_MD5 { 0x00, 0x18 }
+#define GNUTLS_DH_anon_WITH_RIJNDAEL_SHA { 0x00, 0x34 }
 
 #define GNUTLS_DH_DSS_WITH_3DES_EDE_CBC_SHA { 0x00, 0x0D }
 #define GNUTLS_DH_RSA_WITH_3DES_EDE_CBC_SHA { 0x00, 0x10 }
@@ -173,33 +194,29 @@ typedef struct {
 #define GNUTLS_DHE_DSS_WITH_DES_CBC_SHA { 0x00, 0x12 }
 #define GNUTLS_DHE_RSA_WITH_DES_CBC_SHA { 0x00, 0x15 }
 
-#define GNUTLS_RSA_WITH_AES_128_CBC_SHA { 0xFF, 0x2F }
-#define GNUTLS_DH_DSS_WITH_AES_128_CBC_SHA { 0xFF, 0x30 }
-#define GNUTLS_DH_RSA_WITH_AES_128_CBC_SHA { 0xFF, 0x31 }
-#define GNUTLS_DHE_DSS_WITH_AES_128_CBC_SHA { 0xFF, 0x32 }
-#define GNUTLS_DHE_RSA_WITH_AES_128_CBC_SHA { 0xFF, 0x33 }
+#define GNUTLS_RSA_WITH_RIJNDAEL_128_CBC_SHA { 0x00, 0x2F }
+#define GNUTLS_DH_DSS_WITH_RIJNDAEL_128_CBC_SHA { 0x00, 0x30 }
+#define GNUTLS_DH_RSA_WITH_RIJNDAEL_128_CBC_SHA { 0x00, 0x31 }
+#define GNUTLS_DHE_DSS_WITH_RIJNDAEL_128_CBC_SHA { 0x00, 0x32 }
+#define GNUTLS_DHE_RSA_WITH_RIJNDAEL_128_CBC_SHA { 0x00, 0x33 }
      	               
      	               
 static gnutls_cipher_suite_entry cs_algorithms[] = {
-	GNUTLS_CIPHER_SUITE_ENTRY(GNUTLS_DH_anon_WITH_ARCFOUR_MD5,   GNUTLS_ARCFOUR, GNUTLS_KX_ANON_DH, GNUTLS_MAC_MD5),
-	GNUTLS_CIPHER_SUITE_ENTRY(GNUTLS_DH_anon_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_ANON_DH, GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_DSS_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_DH_DSS,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_DSS_WITH_DES_CBC_SHA, 	GNUTLS_DES,  GNUTLS_KX_DH_DSS,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_RSA_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_DH_RSA,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_RSA_WITH_DES_CBC_SHA, 	GNUTLS_DES,  GNUTLS_KX_DH_RSA,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA,GNUTLS_3DES, GNUTLS_KX_DHE_DSS, GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA,GNUTLS_3DES, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_DSS_WITH_DES_CBC_SHA,	GNUTLS_DES,  GNUTLS_KX_DHE_DSS, GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_RSA_WITH_DES_CBC_SHA,	GNUTLS_DES,  GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_ARCFOUR_SHA,         GNUTLS_ARCFOUR, GNUTLS_KX_RSA,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_ARCFOUR_MD5,         GNUTLS_ARCFOUR, GNUTLS_KX_RSA,  GNUTLS_MAC_MD5),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_3DES_EDE_CBC_SHA,    GNUTLS_3DES, GNUTLS_KX_RSA,     GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_DES_CBC_SHA,    	GNUTLS_DES,  GNUTLS_KX_RSA,     GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_AES_128_CBC_SHA,	GNUTLS_AES,  GNUTLS_KX_RSA,	GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_DSS_WITH_AES_128_CBC_SHA,	GNUTLS_AES,  GNUTLS_KX_DH_DSS,  GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_RSA_WITH_AES_128_CBC_SHA,	GNUTLS_AES,  GNUTLS_KX_DH_RSA,	GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_DSS_WITH_AES_128_CBC_SHA,	GNUTLS_AES,  GNUTLS_KX_DHE_DSS,	GNUTLS_MAC_SHA),
-	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_RSA_WITH_AES_128_CBC_SHA,	GNUTLS_AES,  GNUTLS_KX_DHE_RSA,	GNUTLS_MAC_SHA),
+	GNUTLS_CIPHER_SUITE_ENTRY(GNUTLS_DH_anon_WITH_ARCFOUR_MD5,   GNUTLS_ARCFOUR, GNUTLS_KX_ANON_DH, GNUTLS_MAC_MD5, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY(GNUTLS_DH_anon_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_ANON_DH, GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_DSS_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_DH_DSS,  GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_RSA_WITH_3DES_EDE_CBC_SHA, GNUTLS_3DES, GNUTLS_KX_DH_RSA,  GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA,GNUTLS_3DES, GNUTLS_KX_DHE_DSS, GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA,GNUTLS_3DES, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_ARCFOUR_SHA,         GNUTLS_ARCFOUR, GNUTLS_KX_RSA,  GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_ARCFOUR_MD5,         GNUTLS_ARCFOUR, GNUTLS_KX_RSA,  GNUTLS_MAC_MD5, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_3DES_EDE_CBC_SHA,    GNUTLS_3DES, GNUTLS_KX_RSA,     GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_RSA_WITH_RIJNDAEL_128_CBC_SHA,	GNUTLS_RIJNDAEL,  GNUTLS_KX_RSA,	GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_DSS_WITH_RIJNDAEL_128_CBC_SHA,	GNUTLS_RIJNDAEL,  GNUTLS_KX_DH_DSS,  GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DH_RSA_WITH_RIJNDAEL_128_CBC_SHA,	GNUTLS_RIJNDAEL,  GNUTLS_KX_DH_RSA,	GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_DSS_WITH_RIJNDAEL_128_CBC_SHA,	GNUTLS_RIJNDAEL,  GNUTLS_KX_DHE_DSS,	GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY( GNUTLS_DHE_RSA_WITH_RIJNDAEL_128_CBC_SHA,	GNUTLS_RIJNDAEL,  GNUTLS_KX_DHE_RSA,	GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
+	GNUTLS_CIPHER_SUITE_ENTRY(GNUTLS_DH_anon_WITH_RIJNDAEL_SHA,   GNUTLS_RIJNDAEL, GNUTLS_KX_ANON_DH, GNUTLS_MAC_SHA, GNUTLS_COMPRESSION_NULL),
 	{0}
 };
 
@@ -215,7 +232,7 @@ static gnutls_cipher_suite_entry cs_algorithms[] = {
 /* Generic Functions */
 
 /* this function makes the whole string lowercase */
-void tolow(char *str, int size)
+void _gnutls_tolow(char *str, int size)
 {
 	int i;
 
@@ -233,20 +250,15 @@ int _gnutls_mac_get_digest_size(MACAlgorithm algorithm)
 
 }
 
-void _gnutls_mac_set_priority(MACAlgorithm algorithm, int prio)
+int _gnutls_mac_priority(GNUTLS_STATE state, MACAlgorithm algorithm) /* actually returns the priority */
 {
-	GNUTLS_HASH_ALG_LOOP( if (p->priority >= -1) p->priority=prio);
+	int i, num = state->gnutls_internals.MACAlgorithmPriority.algorithms;
+	for (i=0;i<num;i++) {
+		if (state->gnutls_internals.MACAlgorithmPriority.algorithm_priority[i]==algorithm) return i;
+	}
+	return -1;
 }
 
-int _gnutls_mac_priority(MACAlgorithm algorithm) /* actually returns the priority */
-{
-	size_t ret = 0;
-	GNUTLS_HASH_ALG_LOOP(ret = p->priority);
-	return ret;
-
-}
-
-
 char *_gnutls_mac_get_name(MACAlgorithm algorithm)
 {
 	char *ret = NULL;
@@ -258,7 +270,7 @@ char *_gnutls_mac_get_name(MACAlgorithm algorithm)
 
 
 	if (ret != NULL) {
-		tolow(ret, strlen(ret));
+		_gnutls_tolow(ret, strlen(ret));
 		pointerTo_ = strchr(ret, '_');
 
 		while (pointerTo_ != NULL) {
@@ -298,6 +310,67 @@ int _gnutls_mac_is_ok(MACAlgorithm algorithm)
 
 }
 
+/* Compression Functions */
+int _gnutls_compression_priority(GNUTLS_STATE state, CompressionMethod algorithm) /* actually returns the priority */
+{
+	int i, num = state->gnutls_internals.CompressionMethodPriority.algorithms;
+	for (i=0;i<num;i++) {
+		if (state->gnutls_internals.CompressionMethodPriority.algorithm_priority[i]==algorithm) return i;
+	}
+	return -1;
+}
+
+char *_gnutls_compression_get_name(CompressionMethod algorithm)
+{
+	char *ret = NULL;
+	char *pointerTo_;
+
+	/* avoid prefix */
+	GNUTLS_COMPRESSION_ALG_LOOP(ret =
+			     strdup(p->name + sizeof("GNUTLS_") - 1));
+
+
+	if (ret != NULL) {
+		_gnutls_tolow(ret, strlen(ret));
+		pointerTo_ = strchr(ret, '_');
+
+		while (pointerTo_ != NULL) {
+			*pointerTo_ = '-';
+			pointerTo_ = strchr(ret, '_');
+		}
+	}
+	return ret;
+}
+
+int _gnutls_compression_count()
+{
+	uint8 i, counter = 0;
+	char *y;
+
+	for (i = 0; i < 255; i++) {
+		y = _gnutls_compression_get_name(i);
+
+		if (y != NULL) {
+			free(y);
+			counter++;
+		}
+	}
+	return counter;
+}
+
+int _gnutls_compression_is_ok(CompressionMethod algorithm)
+{
+	char *y = _gnutls_compression_get_name(algorithm);
+
+	if (y != NULL) {
+		free(y);
+		return 0;
+	} else {
+		return 1;
+	}
+
+}
+
 
 
 /* CIPHER functions */
@@ -309,18 +382,14 @@ int _gnutls_cipher_get_block_size(BulkCipherAlgorithm algorithm)
 
 }
 
-void _gnutls_cipher_set_priority(BulkCipherAlgorithm algorithm, int prio)
-{
-	GNUTLS_ALG_LOOP(if (p->priority >= -1) p->priority=prio);
-}
-
  /* returns the priority */
-int _gnutls_cipher_priority(BulkCipherAlgorithm algorithm)
+int _gnutls_cipher_priority(GNUTLS_STATE state, BulkCipherAlgorithm algorithm)
 {
-	size_t ret = 0;
-	GNUTLS_ALG_LOOP(ret = p->priority);
-	return ret;
-
+	int i, num = state->gnutls_internals.BulkCipherAlgorithmPriority.algorithms;
+	for (i=0;i<num;i++) {
+		if (state->gnutls_internals.BulkCipherAlgorithmPriority.algorithm_priority[i]==algorithm) return i;
+	}
+	return -1;
 }
 
 
@@ -359,7 +428,7 @@ char *_gnutls_cipher_get_name(BulkCipherAlgorithm algorithm)
 
 
 	if (ret != NULL) {
-		tolow(ret, strlen(ret));
+		_gnutls_tolow(ret, strlen(ret));
 		pointerTo_ = strchr(ret, '_');
 
 		while (pointerTo_ != NULL) {
@@ -410,23 +479,13 @@ int _gnutls_kx_server_certificate(KXAlgorithm algorithm)
 
 }
 
-void _gnutls_kx_set_priority(KXAlgorithm algorithm, int prio)
+int _gnutls_kx_priority(GNUTLS_STATE state, KXAlgorithm algorithm)
 {
-	GNUTLS_KX_ALG_LOOP(if (p->priority >= -1) p->priority=prio);
-}
-
-void _gnutls_prio()
-{
-	GNUTLS_KX_LOOP(fprintf(stderr, "prio: %s/%d\n",p->name, p->priority));
-}
-
-
-int _gnutls_kx_priority(KXAlgorithm algorithm)
-{
-	size_t ret = 0;
-	GNUTLS_KX_ALG_LOOP(ret = p->priority);
-	return ret;
-
+	int i, num = state->gnutls_internals.KXAlgorithmPriority.algorithms;
+	for (i=0;i<num;i++) {
+		if (state->gnutls_internals.KXAlgorithmPriority.algorithm_priority[i]==algorithm) return i;
+	}
+	return -1;
 }
 
 int _gnutls_kx_server_key_exchange(KXAlgorithm algorithm)
@@ -472,7 +531,7 @@ char *_gnutls_kx_get_name(KXAlgorithm algorithm)
 
 
 	if (ret != NULL) {
-		tolow(ret, strlen(ret));
+		_gnutls_tolow(ret, strlen(ret));
 		pointerTo_ = strchr(ret, '_');
 
 		while (pointerTo_ != NULL) {
@@ -521,9 +580,10 @@ int _gnutls_version_cmp(GNUTLS_Version ver1, GNUTLS_Version ver2) {
 	return 0;
 }
 
-int _gnutls_version_is_supported(const GNUTLS_Version version)
+int _gnutls_version_is_supported(GNUTLS_STATE state, const GNUTLS_Version version)
 {
 	size_t ret = 0;
+	/* FIXME: make it to read it from the state */
 	GNUTLS_VERSION_ALG_LOOP(ret = p->supported);
 	return ret;
 }
@@ -555,6 +615,14 @@ MACAlgorithm _gnutls_cipher_suite_get_mac_algo(const GNUTLS_CipherSuite suite)
 
 }
 
+CompressionMethod _gnutls_cipher_suite_get_compression_algo(const GNUTLS_CipherSuite suite)
+{
+	size_t ret = 0;
+	GNUTLS_CIPHER_SUITE_ALG_LOOP(ret = p->compression_algorithm);
+	return ret;
+
+}
+
 char *_gnutls_cipher_suite_get_name(GNUTLS_CipherSuite suite)
 {
 	char *ret = NULL;
@@ -567,7 +635,7 @@ char *_gnutls_cipher_suite_get_name(GNUTLS_CipherSuite suite)
 
 
 	if (ret != NULL) {
-		tolow(ret, strlen(ret));
+		_gnutls_tolow(ret, strlen(ret));
 		pointerTo_ = strchr(ret, '_');
 
 		while (pointerTo_ != NULL) {
@@ -612,13 +680,33 @@ int _gnutls_cipher_suite_count()
 	return counter;
 }
 
+static void bsort(GNUTLS_STATE state, void *_base, size_t nmemb, size_t size, int (*compar)(GNUTLS_STATE, const void *, const void *)) {
+int i,j;
+int full=nmemb*size;
+char* base=_base;
+char* tmp=gnutls_malloc(size);
+
+	for (i=0;i<full;i+=size) {
+		for (j=0;j<full;j+=size) {
+			if (compar(state, &base[i], &base[j]) < 0) {
+				memcpy(tmp, &base[i], size);
+				memcpy(&base[i], &base[j], size);
+				memcpy(&base[j], tmp, size);
+			}			
+		}
+	}
+	free(tmp);
+
+}
+
+
 /* a compare function for hash(mac) algorithms (using priorities). For use with qsort */
-static int _gnutls_compare_mac_algo(const void* i_A1, const void* i_A2)
+static int _gnutls_compare_mac_algo(GNUTLS_STATE state, const void* i_A1, const void* i_A2)
 {
 	MACAlgorithm A1 = _gnutls_cipher_suite_get_mac_algo( *(GNUTLS_CipherSuite*)i_A1);
 	MACAlgorithm A2 = _gnutls_cipher_suite_get_mac_algo( *(GNUTLS_CipherSuite*)i_A2);
-	int p1 = _gnutls_mac_priority(A1);
-	int p2 = _gnutls_mac_priority(A2);
+	int p1 = _gnutls_mac_priority(state, A1);
+	int p2 = _gnutls_mac_priority(state, A2);
 
 	if (p1 > p2) {
 		return -1;
@@ -634,12 +722,12 @@ static int _gnutls_compare_mac_algo(const void* i_A1, const void* i_A2)
 
 
 /* a compare function for block algorithms (using priorities). For use with qsort */
-static int _gnutls_compare_cipher_algo(const void* i_A1, const void* i_A2)
+static int _gnutls_compare_cipher_algo(GNUTLS_STATE state, const void* i_A1, const void* i_A2)
 {
 	BulkCipherAlgorithm A1 = _gnutls_cipher_suite_get_cipher_algo( *(GNUTLS_CipherSuite*)i_A1);
 	BulkCipherAlgorithm A2 = _gnutls_cipher_suite_get_cipher_algo( *(GNUTLS_CipherSuite*)i_A2);
-	int p1 = _gnutls_cipher_priority(A1);
-	int p2 = _gnutls_cipher_priority(A2);
+	int p1 = _gnutls_cipher_priority(state, A1);
+	int p2 = _gnutls_cipher_priority(state, A2);
 
 	if (p1 > p2) {
 		return -1; /* we actually want descending order */
@@ -655,12 +743,12 @@ static int _gnutls_compare_cipher_algo(const void* i_A1, const void* i_A2)
 
 
 /* a compare function for KX algorithms (using priorities). For use with qsort */
-static int _gnutls_compare_kx_algo(const void* i_A1, const void* i_A2)
+static int _gnutls_compare_kx_algo(GNUTLS_STATE state, const void* i_A1, const void* i_A2)
 {
 	KXAlgorithm A1 = _gnutls_cipher_suite_get_kx_algo(*(GNUTLS_CipherSuite*)i_A1);
 	KXAlgorithm A2 = _gnutls_cipher_suite_get_kx_algo(*(GNUTLS_CipherSuite*)i_A2);
-	int p1 = _gnutls_kx_priority(A1);
-	int p2 = _gnutls_kx_priority(A2);
+	int p1 = _gnutls_kx_priority(state, A1);
+	int p2 = _gnutls_kx_priority(state, A2);
 
 	if (p1 > p2) {
 		return -1;
@@ -674,7 +762,7 @@ static int _gnutls_compare_kx_algo(const void* i_A1, const void* i_A2)
 	}
 }
 
-int _gnutls_supported_ciphersuites(GNUTLS_CipherSuite ** ciphers)
+int _gnutls_supported_ciphersuites(GNUTLS_STATE state, GNUTLS_CipherSuite ** ciphers)
 {
 
 	int i, ret_count, j=0;
@@ -698,23 +786,22 @@ int _gnutls_supported_ciphersuites(GNUTLS_CipherSuite ** ciphers)
 	}
 
 /* First sort using MAC priority (lowest) */
-	qsort(tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_mac_algo);
+	bsort(state, tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_mac_algo);
 
 /* then sort using block algorithm's priorities */
-	qsort(tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_cipher_algo);
+	bsort(state, tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_cipher_algo);
 
-/* Last try KX algorithms priority */
-	qsort(tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_kx_algo);
+/* Last try KX algorithms priority (highest) */
+	bsort(state, tmp_ciphers, count, sizeof(GNUTLS_CipherSuite), _gnutls_compare_kx_algo);
 
 	for (i = 0; i < count; i++) {
 
-		if (_gnutls_kx_priority( _gnutls_cipher_suite_get_kx_algo(tmp_ciphers[i])) < 0) continue;
-		if (_gnutls_mac_priority( _gnutls_cipher_suite_get_mac_algo(tmp_ciphers[i])) < 0) continue;
-		if (_gnutls_cipher_priority( _gnutls_cipher_suite_get_cipher_algo(tmp_ciphers[i])) < 0) continue;
+		if (_gnutls_kx_priority( state, _gnutls_cipher_suite_get_kx_algo(tmp_ciphers[i])) < 0) continue;
+		if (_gnutls_mac_priority( state, _gnutls_cipher_suite_get_mac_algo(tmp_ciphers[i])) < 0) continue;
+		if (_gnutls_cipher_priority( state, _gnutls_cipher_suite_get_cipher_algo(tmp_ciphers[i])) < 0) continue;
 
 		(*ciphers)[j].CipherSuite[0] = tmp_ciphers[i].CipherSuite[0];
 		(*ciphers)[j].CipherSuite[1] = tmp_ciphers[i].CipherSuite[1];
-/*		fprintf(stderr, "%d: %s\n", j, _gnutls_cipher_suite_get_name((*ciphers)[j])); */
 		j++;
 	}
 	ret_count=j;
@@ -732,3 +819,18 @@ int _gnutls_supported_ciphersuites(GNUTLS_CipherSuite ** ciphers)
 	gnutls_free(tmp_ciphers);
 	return ret_count;
 }
+
+/* For compression  */
+#define SUPPORTED_COMPRESSION_METHODS state->gnutls_internals.CompressionMethodPriority.algorithms
+int _gnutls_supported_compression_methods(GNUTLS_STATE state, CompressionMethod ** comp)
+{
+int i;
+ 	*comp =
+            gnutls_malloc(SUPPORTED_COMPRESSION_METHODS * 1);
+
+	for (i=0;i<SUPPORTED_COMPRESSION_METHODS;i++) {
+		(*comp)[i] = state->gnutls_internals.CompressionMethodPriority.algorithm_priority[i];
+	}
+
+	return SUPPORTED_COMPRESSION_METHODS;
+}
