@@ -108,7 +108,7 @@ int gnutls_getDataFromBuffer(ContentType type, GNUTLS_STATE state, char *data, i
 	return length;
 }
 
-ssize_t Read(int fd, void *iptr, size_t sizeOfPtr)
+ssize_t _gnutls_Read(int fd, void *iptr, size_t sizeOfPtr)
 {
 	size_t left;
 	ssize_t i=0;
@@ -151,7 +151,7 @@ ssize_t Read(int fd, void *iptr, size_t sizeOfPtr)
 }
 
 
-ssize_t Write(int fd, const void *iptr, size_t n)
+ssize_t _gnutls_Write(int fd, const void *iptr, size_t n)
 {
 	size_t left;
 #ifdef WRITE_DEBUG
@@ -226,3 +226,76 @@ ssize_t _gnutls_Recv_int(int fd, GNUTLS_STATE state, ContentType type, void *ipt
 
 	return (sizeOfPtr - left);
 }
+
+int gnutls_insertHashDataBuffer( GNUTLS_STATE state, char *data, int length)
+{
+	int old_buffer;
+	
+	old_buffer = state->gnutls_internals.hash_bufferSize;
+
+	state->gnutls_internals.hash_bufferSize += length;
+#ifdef HANDSHAKE_DEBUG
+	fprintf(stderr, "Inserted %d bytes of Hash Data into buffer\n", length);
+#endif
+	state->gnutls_internals.hash_buffer =
+		    gnutls_realloc(state->gnutls_internals.hash_buffer,
+			   state->gnutls_internals.hash_bufferSize);
+		memmove(&state->gnutls_internals.hash_buffer[old_buffer], data, length);
+
+	return 0;
+}
+
+int gnutls_getHashDataBufferSize( GNUTLS_STATE state)
+{
+
+	return state->gnutls_internals.hash_bufferSize;
+}
+
+int gnutls_getHashDataFromBuffer( GNUTLS_STATE state, char *data, int length)
+{
+	if (length > state->gnutls_internals.hash_bufferSize) {
+		length = state->gnutls_internals.hash_bufferSize;
+	}
+#ifdef HANDSHAKE_DEBUG
+	fprintf(stderr, "Got %d bytes of Hash Data from buffer\n", length);
+#endif
+	state->gnutls_internals.hash_bufferSize -= length;
+	memmove(data, state->gnutls_internals.hash_buffer, length);
+	/* overwrite buffer */
+	memmove(state->gnutls_internals.hash_buffer,
+		&state->gnutls_internals.hash_buffer[length],
+		state->gnutls_internals.hash_bufferSize);
+	state->gnutls_internals.hash_buffer =
+	    gnutls_realloc(state->gnutls_internals.hash_buffer,
+			   state->gnutls_internals.hash_bufferSize);
+
+	return length;	
+
+}
+
+int gnutls_readHashDataFromBuffer( GNUTLS_STATE state, char *data, int length)
+{
+	if (length > state->gnutls_internals.hash_bufferSize) {
+		length = state->gnutls_internals.hash_bufferSize;
+	}
+#ifdef HANDSHAKE_DEBUG
+	fprintf(stderr, "Read %d bytes of Hash Data from buffer\n", length);
+#endif
+	memmove(data, state->gnutls_internals.hash_buffer, length);
+	return length;	
+}
+
+
+
+int gnutls_clearHashDataBuffer( GNUTLS_STATE state)
+{
+
+#ifdef HANDSHAKE_DEBUG
+	fprintf(stderr, "Cleared Hash Data from buffer\n");
+#endif
+	state->gnutls_internals.hash_bufferSize = 0;
+	gnutls_free(state->gnutls_internals.hash_buffer);
+	state->gnutls_internals.hash_buffer = NULL;
+	
+	return 0;
+}
