@@ -31,14 +31,7 @@ GNUTLS_CIPHER_HANDLE ret;
 	case GNUTLS_NULL:
 		ret = GNUTLS_CIPHER_FAILED;
 		break;
-	case GNUTLS_DES:
-#ifdef USE_MCRYPT
-		ret = mcrypt_module_open( "des", NULL, "cbc", NULL);
-#else
-		ret = GNUTLS_CIPHER_FAILED;
-#endif
-		break;
-	case GNUTLS_AES:
+	case GNUTLS_RIJNDAEL:
 #ifdef USE_MCRYPT
 		ret = mcrypt_module_open( "rijndael-128", NULL, "cbc", NULL);
 #else
@@ -56,7 +49,7 @@ GNUTLS_CIPHER_HANDLE ret;
 #ifdef USE_MCRYPT
 		ret = mcrypt_module_open( "arcfour", NULL, "stream", NULL);
 #else
-		ret = GNUTLS_CIPHER_FAILED;
+		ret = gcry_cipher_open(GCRY_CIPHER_ARCFOUR, GCRY_CIPHER_MODE_STREAM, 0);
 #endif
 		break;
 	default:
@@ -70,7 +63,7 @@ GNUTLS_CIPHER_HANDLE ret;
 		}; 
 #else
 		gcry_cipher_setkey(ret, key, keysize);
-		gcry_cipher_setiv(ret, iv, ivsize);
+		if (iv!=NULL && ivsize>0) gcry_cipher_setiv(ret, iv, ivsize);
 #endif
 	}
 
