@@ -2,6 +2,7 @@
 
 #define GNUTLS_INT_H
 
+//#define HANDSHAKE_DEBUG
 //#define HARD_DEBUG
 //#define READ_DEBUG
 //#define WRITE_DEBUG
@@ -18,14 +19,14 @@
 #define gnutls_mpi_release mpi_release
 
 #define svoid void /* for functions that allocate using secure_free */
-#define secure_free(x) if (x!=NULL) free(x)
+#define secure_free free
 #define secure_malloc malloc
 #define secure_realloc realloc
 #define secure_calloc calloc
 #define gnutls_malloc malloc
 #define gnutls_realloc realloc
 #define gnutls_calloc calloc
-#define gnutls_free(x) if (x!=NULL) free(x)
+#define gnutls_free free
 
 typedef struct {
 	uint8	pint[3];
@@ -82,13 +83,13 @@ typedef struct {
 
 /* STATE */
 enum ConnectionEnd { GNUTLS_SERVER, GNUTLS_CLIENT };
-enum BulkCipherAlgorithm { GNUTLS_NULL, GNUTLS_ARCFOUR=1, GNUTLS_DES=3, GNUTLS_3DES = 4, GNUTLS_AES };
+enum BulkCipherAlgorithm { GNUTLS_NULL, GNUTLS_ARCFOUR=1, GNUTLS_3DES = 4, GNUTLS_RIJNDAEL };
 enum KXAlgorithm { GNUTLS_KX_RSA, GNUTLS_KX_DHE_DSS, GNUTLS_KX_DHE_RSA, GNUTLS_KX_DH_DSS, GNUTLS_KX_DH_RSA, GNUTLS_KX_ANON_DH };
 enum KeyExchangeAlgorithm { GNUTLS_RSA, GNUTLS_DIFFIE_HELLMAN };
 enum CipherType { CIPHER_STREAM, CIPHER_BLOCK };
 enum IsExportable { EXPORTABLE_TRUE, EXPORTABLE_FALSE };
 enum MACAlgorithm { GNUTLS_MAC_NULL, GNUTLS_MAC_MD5, GNUTLS_MAC_SHA };
-enum CompressionMethod { COMPRESSION_NULL };
+enum CompressionMethod { GNUTLS_COMPRESSION_NULL, GNUTLS_ZLIB=224 };
 
 enum ValidSession { VALID_TRUE, VALID_FALSE };
 enum ResumableSession { RESUME_TRUE, RESUME_FALSE };
@@ -111,8 +112,7 @@ typedef struct {
 	ConnectionEnd entity;
 	BulkCipherAlgorithm bulk_cipher_algorithm;
 	CipherType cipher_type;
-	uint8 IV_size;   /* not specified in the protocol, but later it
-			  * uses it */
+	uint8 IV_size;
 	uint8 key_size;
 	uint8 key_material_length;
 	IsExportable is_exportable;
@@ -145,13 +145,13 @@ extern GNUTLS_Version GNUTLS_SSL3;
 
 typedef struct {
 	GNUTLS_Version version;
-	opaque* read_compression_state;
-	opaque* write_compression_state;
+	opaque* 	read_compression_state;
+	opaque* 	write_compression_state;
 	GNUTLS_CIPHER_HANDLE write_cipher_state;
 	GNUTLS_CIPHER_HANDLE read_cipher_state;
-	opaque* read_mac_secret;
-	opaque* write_mac_secret;
-	uint8   mac_secret_size;
+	opaque* 	read_mac_secret;
+	opaque* 	write_mac_secret;
+	uint8   	mac_secret_size;
 	uint64	read_sequence_number;
 	uint64	write_sequence_number;
 } ConnectionState;
@@ -160,36 +160,50 @@ typedef struct {
 	uint8 CipherSuite[2];
 } GNUTLS_CipherSuite;
 
+typedef struct {
+	int* algorithm_priority;
+	int algorithms;
+} BulkCipherAlgorithm_Priority;
+
+typedef struct {
+	int* algorithm_priority;
+	int algorithms;
+} MACAlgorithm_Priority;
+
+typedef struct {
+	int* algorithm_priority;
+	int algorithms;
+} KXAlgorithm_Priority;
+
+typedef struct {
+	int* algorithm_priority;
+	int algorithms;
+} CompressionMethod_Priority;
 
 typedef struct {
 	char*			buffer;
 	uint32			bufferSize;
+	char*			hash_buffer; /* used in SSL3 */
+	uint32			hash_bufferSize; /* used in SSL3 */
 	char*			buffer_handshake;
 	uint32			bufferSize_handshake;
-	ResumableSession	resumable;
-	ValidSession		valid_connection;
-	AlertDescription	last_alert;
+	ResumableSession	resumable; /* TRUE or FALSE */
+	ValidSession		valid_connection; /* true or FALSE */
+	AlertDescription	last_alert; /* last alert received */
 	GNUTLS_CipherSuite	current_cipher_suite;
 	CompressionMethod	compression_method;
-	/* for the handshake protocol */
-	GNUTLS_HASH_HANDLE	client_td_md5;
-	GNUTLS_HASH_HANDLE	client_td_sha1;
-	void*			client_md_md5;
-	void*			client_md_sha1;
-	GNUTLS_HASH_HANDLE	server_td_md5;
-	GNUTLS_HASH_HANDLE	server_td_sha1;
-	void*			server_md_md5;
-	void*			server_md_sha1;
-	int			server_hash;
-	int			client_hash;
 	/* For DH KX */
-	MPI			KEY;
-	MPI			client_Y;
-	MPI			client_g;
-	MPI			client_p;
-	MPI			dh_secret;
-	int			certificate_requested;
-	int			certificate_verify_needed;
+	MPI				KEY;
+	MPI				client_Y;
+	MPI				client_g;
+	MPI				client_p;
+	MPI				dh_secret;
+	int				certificate_requested; /* non zero if client certificate was requested */
+	int				certificate_verify_needed; /* non zero if we should expect for certificate verify */
+	BulkCipherAlgorithm_Priority	BulkCipherAlgorithmPriority;
+	MACAlgorithm_Priority		MACAlgorithmPriority;
+	KXAlgorithm_Priority		KXAlgorithmPriority;
+	CompressionMethod_Priority	CompressionMethodPriority;
 } GNUTLS_INTERNALS;
 
 typedef struct {
@@ -291,11 +305,10 @@ int gnutls_close(int cd, GNUTLS_STATE state);
 svoid *gnutls_PRF( opaque * secret, int secret_size, uint8 * label,
 		  int label_size, opaque * seed, int seed_size,
 		  int total_bytes);
-int _gnutls_valid_version( GNUTLS_STATE state, int major, int minor);
 void gnutls_set_current_version(GNUTLS_STATE state, GNUTLS_Version version);
 GNUTLS_Version gnutls_get_current_version(GNUTLS_STATE state);
 int _gnutls_set_keys(GNUTLS_STATE state);
-ssize_t gnutls_send_int(int cd, GNUTLS_STATE state, ContentType type, char* data, size_t sizeofdata);
+ssize_t gnutls_send_int(int cd, GNUTLS_STATE state, ContentType type, void* data, size_t sizeofdata);
 ssize_t gnutls_recv_int(int cd, GNUTLS_STATE state, ContentType type, char* data, size_t sizeofdata);
 int _gnutls_send_change_cipher_spec(int cd, GNUTLS_STATE state);
 int _gnutls_version_cmp(GNUTLS_Version ver1, GNUTLS_Version ver2);
